/* --------------------------------------------------------------------- **
**                                                                       **
** By default we use mmap to map memory.                                 **
** We can also use shared memory segments (shm) to map memory.           **
** To map memory using shm we should:                                    **
**   define the macro USE_SHM_FOR_SHARED_MEMORY in this file             **
**   or                                                                  **
**   add -DUSE_SHM_FOR_SHARED_MEMORY to the Makefile YAP_EXTRAS variable **
**                                                                       **
** --------------------------------------------------------------------- */

#define USE_SHM_FOR_SHARED_MEMORY



/* ----------------------------------------------- **
**      Includes, defines and local variables      **
** ----------------------------------------------- */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include "alloc.h"
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/shm.h>
#include <sys/mman.h>


#define K                          ((Int) 1024) /* as in 'alloc.c' */
#define MinHGap                    256 * K  /* as in 'alloc.c' */
#define KBYTES                     K
#define OPT_CHUNK_SIZE             5000 * KBYTES
#define ALIGN	                   3
#define ALIGNMASK                  0xfffffffc
#define ADJUST_SIZE(X)             ((X + ALIGN) & ALIGNMASK)
#define PAGESIZE                   sysconf(_SC_PAGESIZE)
#define ADJUST_SIZE_TO_PAGE(size)  (size - (size) % PAGESIZE + PAGESIZE)

#ifdef YAPOR
#define OR_FRAMES_ALLOC_NUMBER 50
#define OR_FRAME_SIZE ADJUST_SIZE(sizeof(struct or_frame))
#define ANSWER_FRAMES_ALLOC_NUMBER 50
#define ANSWER_FRAME_SIZE ADJUST_SIZE(sizeof(struct answer_frame))
#define SOLUTION_FRAMES_ALLOC_NUMBER 50
#define SOLUTION_FRAME_SIZE ADJUST_SIZE(sizeof(struct solution_frame))
#endif /* YAPOR */
#ifdef TABLING
#define TABLE_ENTRIES_ALLOC_NUMBER 10
#define TABLE_ENTRY_SIZE ADJUST_SIZE(sizeof(struct table_entry))
#define SUBGOAL_TRIE_NODES_ALLOC_NUMBER 50
#define SUBGOAL_TRIE_NODE_SIZE ADJUST_SIZE(sizeof(struct subgoal_trie_node))
#define SUBGOAL_FRAMES_ALLOC_NUMBER 50
#define SUBGOAL_FRAME_SIZE ADJUST_SIZE(sizeof(struct subgoal_frame))
#define ANSWER_TRIE_NODES_ALLOC_NUMBER 50
#define ANSWER_TRIE_NODE_SIZE ADJUST_SIZE(sizeof(struct answer_trie_node))
#define DEPENDENCY_FRAMES_ALLOC_NUMBER 50
#define DEPENDENCY_FRAME_SIZE ADJUST_SIZE(sizeof(struct dependency_frame))
#ifdef YAPOR
#define SUSPENSION_FRAMES_ALLOC_NUMBER 20
#define SUSPENSION_FRAME_SIZE ADJUST_SIZE(sizeof(struct suspension_frame))
#endif /* YAPOR */
#endif /* TABLING */


#ifdef USE_SHM_FOR_SHARED_MEMORY
int shm_mapid[MAX_WORKERS + 2];
#else
int fd_mapfile;
#endif /* USE_SHM_FOR_SHARED_MEMORY */



/* --------------------------- **
**      Global functions       **
** --------------------------- */

long global_data_size(void) {
  return ADJUST_SIZE(sizeof(struct global_data));
}


long local_data_size(void) {
  return ADJUST_SIZE(sizeof(struct local_data));
}


void open_mapfile(void) {
#ifndef USE_SHM_FOR_SHARED_MEMORY
  char mapfile[20];
  strcpy(mapfile,"/tmp/mapfile");
  itos(getpid(), &mapfile[12]);
  if ((fd_mapfile = open(mapfile, O_RDWR|O_CREAT|O_TRUNC, 0666)) < 0)
    abort_optyap("open error in function open_mapfile");
#endif /* USE_SHM_FOR_SHARED_MEMORY */
  return;
}


void close_mapfile(void) {
#ifndef USE_SHM_FOR_SHARED_MEMORY
  close(fd_mapfile);
#endif /* USE_SHM_FOR_SHARED_MEMORY */
  return;
}


void map_memory(long HeapArea, long GlobalLocalArea, long TrailAuxArea, int n_workers) {
#ifdef YAPOR
  int i;
#endif /* YAPOR */
  long WorkerArea;
  long OPTArea;
  long TotalArea;
  void *mmap_addr;
#ifdef ACOW
  void *private_fd_mapfile;
  void *private_mmap_addr;
#endif /* ACOW */

  HeapArea = ADJUST_SIZE_TO_PAGE(HeapArea * KBYTES  + global_data_size());
  GlobalLocalArea = ADJUST_SIZE(GlobalLocalArea * KBYTES);
  TrailAuxArea = ADJUST_SIZE(TrailAuxArea * KBYTES);
#ifdef ACOW
  WorkerArea = ADJUST_SIZE_TO_PAGE(local_data_size());
#else
  WorkerArea = ADJUST_SIZE_TO_PAGE(local_data_size() + GlobalLocalArea + TrailAuxArea);
#endif /* ACOW */
  OPTArea = ADJUST_SIZE_TO_PAGE(OPT_CHUNK_SIZE);
  TotalArea = HeapArea + WorkerArea * n_workers + OPTArea;

#ifdef USE_SHM_FOR_SHARED_MEMORY
  /* mapping heap area */
  if((shm_mapid[n_workers] = shmget(IPC_PRIVATE, (size_t)HeapArea, SHM_R|SHM_W)) == -1) 
    abort_optyap("shmget error in function map_memory");
  if((mmap_addr = shmat(shm_mapid[n_workers], (void *) MMAP_ADDR, 0)) == (void *) -1)
    abort_optyap("shmat error in function map_memory");
#else
  /* mapping total area */
  open_mapfile();
  if (lseek(fd_mapfile, TotalArea, SEEK_SET) < 0) 
    abort_optyap("lseek error in function map_memory");
  if (write(fd_mapfile, "", 1) < 0) 
    abort_optyap("write error in function map_memory");
  if ((mmap_addr = mmap((void *) MMAP_ADDR, (size_t) TotalArea, PROT_READ|PROT_WRITE, 
                         MAP_SHARED|MAP_FIXED, fd_mapfile, 0)) == (void *) -1)
    abort_optyap("mmap error in function map_memory");
#endif /* USE_SHM_FOR_SHARED_MEMORY */

#ifdef YAPOR
  for (i = 0; i < n_workers; i++) {
    /* initialize worker vars */
    worker_area(i) = mmap_addr + HeapArea + i * WorkerArea;
    worker_offset(i) = worker_area(i) - worker_area(0);
    REMOTE[i] = (struct local_data *) worker_area(i);
#ifdef USE_SHM_FOR_SHARED_MEMORY
    /* mapping worker area */
    if((shm_mapid[i] = shmget(IPC_PRIVATE, (size_t)WorkerArea, SHM_R|SHM_W)) == -1) 
      abort_optyap("shmget error in function map_memory");
    if(shmat(shm_mapid[i], worker_area(i), 0) == (void *) -1)
      abort_optyap("shmat error in function map_memory");
#endif /* USE_SHM_FOR_SHARED_MEMORY */
  }
#else /* TABLING */
#ifdef USE_SHM_FOR_SHARED_MEMORY
  /* mapping worker area */
  if((shm_mapid[0] = shmget(IPC_PRIVATE, (size_t)WorkerArea, SHM_R|SHM_W)) == -1) 
    abort_optyap("shmget error in function map_memory");
  if(shmat(shm_mapid[0], mmap_addr + HeapArea, 0) == (void *) -1)
    abort_optyap("shmat error in function map_memory");
#endif /* USE_SHM_FOR_SHARED_MEMORY */
#endif /* YAPOR */

  /* initialize global vars */
  GLOBAL = (struct global_data *) (mmap_addr + HeapArea - global_data_size());
  BaseWorkArea = mmap_addr;
  TopWorkArea = BaseWorkArea + TotalArea;
  TopAllocArea = BaseWorkArea + HeapArea + n_workers * WorkerArea;
#if defined(YAPOR) && defined(TABLING)
  SuspensionArea = TopWorkArea;
#endif /* YAPOR && TABLING */
#ifdef USE_SHM_FOR_SHARED_MEMORY
  /* mapping opt area */
  if ((shm_mapid[n_workers + 1] = shmget(IPC_PRIVATE, (size_t)OPTArea, SHM_R|SHM_W)) == -1)
    abort_optyap("shmget error in function map_memory");
  if (shmat(shm_mapid[n_workers + 1], TopAllocArea, 0) == (void *) -1)
    abort_optyap("shmat error in function map_memory");
#endif /* USE_SHM_FOR_SHARED_MEMORY */
#ifdef ACOW
  /* alloc space for private stacks */
  if ((private_fd_mapfile = open("/dev/zero", O_RDWR)) < 0)
    abort_optyap("open error in function map_memory");
  private_mmap_addr = mmap_addr + TotalArea;
  if (mmap(private_mmap_addr, GlobalLocalArea + TrailAuxArea, PROT_READ|PROT_WRITE, 
           MAP_PRIVATE|MAP_FIXED, private_fd_mapfile, 0) == (void *) -1)
    abort_optyap("mmap error in function map_memory");
  close(private_fd_mapfile);
#endif /* ACOW */

  LOCAL = (struct local_data *) (mmap_addr + HeapArea);
#ifdef ACOW 
  GlobalBase = private_mmap_addr;
#else
  GlobalBase = mmap_addr + HeapArea + local_data_size();
#endif /* ACOW */
#ifdef SBA
  /* alloc space for the sparse binding array */
  sba_size = WorkerArea * n_workers;
  if ((binding_array = (char *)malloc(sba_size)) == NULL)
    abort_optyap("malloc error in function map_memory");
  sba_offset = binding_array - GlobalBase;
  sba_end = (int)binding_array + sba_size;
#endif /* SBA */
  TrailBase = GlobalBase + GlobalLocalArea;
  LocalBase = TrailBase - CellSize;
  if (TrailAuxArea > 262144)                       /* 262144 = 256 * 1024 */ 
    TrailTop = TrailBase + TrailAuxArea - 131072;  /* 131072 = 262144 / 2 */ 
  else
    TrailTop = TrailBase + TrailAuxArea / 2;
  AuxTop = TrailBase + TrailAuxArea - CellSize;
  AuxSp = (CELL *) AuxTop;
  YAP_InitHeap((void *)BaseWorkArea);
#if USE_OFFSETS
  AtomBase = HeapBase;
#endif

  return;
}


void unmap_memory (void) {
#ifdef USE_SHM_FOR_SHARED_MEMORY
  int i;
#else
  char MapFile[20];
#endif /* USE_SHM_FOR_SHARED_MEMORY */

#ifdef YAPOR
  if (GLOBAL) {
    int proc;
    for (proc = 0; proc < number_workers; proc++) {
      if (proc != worker_id && worker_pid(proc) != 0) {
        if (kill(worker_pid(proc), SIGKILL) != 0)
          INFORMATION_MESSAGE("Can't kill process %d", worker_pid(proc));
        else 
          INFORMATION_MESSAGE("Killing process %d", worker_pid(proc));
      }
    }
#ifdef ACOW
    if (kill(GLOBAL_master_worker, SIGINT) != 0)
      INFORMATION_MESSAGE("Can't kill process %d", GLOBAL_master_worker);
    else 
      INFORMATION_MESSAGE("Killing process %d", GLOBAL_master_worker);
#endif /* ACOW */
  }
#endif /* YAPOR */

#ifdef USE_SHM_FOR_SHARED_MEMORY
#ifdef YAPOR
  for (i = 0; i < number_workers + 2; i++) {
#else /* TABLING */
  for (i = 0; i < 1 + 2; i++) {
#endif /* YAPOR */
    if (shmctl(shm_mapid[i], IPC_RMID, 0) == 0)
      INFORMATION_MESSAGE("Removing shared memory segment %d", shm_mapid[i]);
    else INFORMATION_MESSAGE("Can't remove shared memory segment %d", shm_mapid[i]);
  }
#else
  strcpy(MapFile,"/tmp/mapfile");
#ifdef YAPOR
#ifdef ACOW
  itos(GLOBAL_master_worker, &MapFile[12]);
#else /* ENV_COPY || SBA */
  itos(worker_pid(0), &MapFile[12]);
#endif
#else /* TABLING */
  itos(getpid(), &MapFile[12]);
#endif /* YAPOR */
  if (remove(MapFile) == 0)
    INFORMATION_MESSAGE("Removing mapfile \"%s\"", MapFile);
  else INFORMATION_MESSAGE("Can't remove mapfile \"%s\"", MapFile);
#endif /* USE_SHM_FOR_SHARED_MEMORY */
  return;
}


void alloc_all_optyap_structs(void) {
#ifdef YAPOR
  alloc_or_frames();
  alloc_answer_frames();
  alloc_solution_frames();
#endif /* YAPOR */
#ifdef TABLING
  alloc_table_entries();
  alloc_subgoal_trie_nodes();
  alloc_subgoal_frames();
  alloc_answer_trie_nodes();
  alloc_dependency_frames();
#ifdef YAPOR
  alloc_suspension_frames();
#endif /* YAPOR */
#endif /* TABLING */
  return;
}


#ifdef YAPOR
void remap_memory(void) {
#ifdef ACOW
  /* do nothing */
#endif /* ACOW */
#ifdef SBA
  /* setup workers so that they have different areas */
  long WorkerArea = worker_offset(1);

  GlobalBase += worker_id * WorkerArea;
  TrailBase += worker_id * WorkerArea;
  LocalBase += worker_id * WorkerArea;
  TrailTop += worker_id * WorkerArea;
  AuxTop += worker_id * WorkerArea;
  AuxSp = (CELL *) AuxTop; 
#endif /* SBA */
#ifdef ENV_COPY
  void *remap_addr;
  long remap_offset;
  long WorkerArea;
  int i;

  remap_addr = worker_area(0);
  remap_offset = remap_addr - BaseWorkArea;
  WorkerArea = worker_offset(1);
#ifdef USE_SHM_FOR_SHARED_MEMORY
  for (i = 0; i < number_workers; i++) {
    if (shmdt(worker_area(i)) == -1)
      abort_optyap("shmdt error in function remap_memory");
  }
  for (i = 0; i < number_workers; i++) {
    worker_area(i) = remap_addr + ((number_workers + i - worker_id) % number_workers) * WorkerArea;
    if(shmat(shm_mapid[i], worker_area(i), 0) == (void *) -1)
      abort_optyap("shmat error in function remap_memory");
  }
#else
  if (munmap(remap_addr, (size_t)(WorkerArea * number_workers)) == -1)
    abort_optyap("munmap error in function remap_memory");
  for (i = 0; i < number_workers; i++) {
    worker_area(i) = remap_addr + ((number_workers + i - worker_id) % number_workers) * WorkerArea;
    if (mmap(worker_area(i), (size_t)WorkerArea, PROT_READ|PROT_WRITE, 
        MAP_SHARED|MAP_FIXED, fd_mapfile, remap_offset + i * WorkerArea) == (void *) -1)
      abort_optyap("mmap error in function remap_memory");
  }
#endif /* USE_SHM_FOR_SHARED_MEMORY */
  for (i = 0; i < number_workers; i++) {
    worker_offset(i) = worker_area(i) - worker_area(worker_id);
    REMOTE[i] = (struct local_data *) worker_area(i);
  }
#endif /* ENV_COPY */
  return;
}


void alloc_or_frames(void) {
  int i;
  or_fr_ptr aux;

  YAPOR_DEBUG_MESSAGE("Allocating %d or frames", OR_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef TABLING
  if (OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_or_frames)");
  GLOBAL_next_free_or_fr = TopAllocArea;
  TopAllocArea += OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_or_fr;
  for (i = 1; i < OR_FRAMES_ALLOC_NUMBER; i++) {
    OrFr_next(aux) = (or_fr_ptr) ((unsigned int)aux + OR_FRAME_SIZE);
    aux = OrFr_next(aux);
  }
  OrFr_next(aux) = NULL;
#ifdef YAPOR_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_or_fr += OR_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* YAPOR_STATISTICS */ 
  return;
}


void alloc_answer_frames(void) {
  int i;
  ans_fr_ptr aux;

  YAPOR_DEBUG_MESSAGE("Allocating %d answer frames", ANSWER_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef TABLING
  if (ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_answer_frames)");
  GLOBAL_next_free_ans_fr = TopAllocArea;
  TopAllocArea += ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_ans_fr;
  for (i = 1; i < ANSWER_FRAMES_ALLOC_NUMBER; i++) {
    AnsFr_next(aux) = (ans_fr_ptr) ((unsigned int)aux + ANSWER_FRAME_SIZE);
    aux = AnsFr_next(aux);
  }
  AnsFr_next(aux) = NULL;
#ifdef YAPOR_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_ans_fr += ANSWER_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* YAPOR_STATISTICS */ 
  return;
}


void alloc_solution_frames(void) {
  int i;
  sol_fr_ptr aux;

  YAPOR_DEBUG_MESSAGE("Allocating %d solution frames", SOLUTION_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef TABLING
  if (SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_solution_frames)");
  GLOBAL_next_free_sol_fr = TopAllocArea;
  TopAllocArea += SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_sol_fr;
  for (i = 1; i < SOLUTION_FRAMES_ALLOC_NUMBER; i++) {
    SolFr_next(aux) = (sol_fr_ptr) ((unsigned int)aux + SOLUTION_FRAME_SIZE);
    aux = SolFr_next(aux);
  }
  SolFr_next(aux) = NULL;
#ifdef YAPOR_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_sol_fr += SOLUTION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* YAPOR_STATISTICS */ 
  return;
}
#endif /* YAPOR */


#ifdef TABLING
void alloc_table_entries(void) {
  int i;
  tab_ent_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d table entries", TABLE_ENTRIES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef YAPOR
  if (TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_table_entries)");
  GLOBAL_next_free_tab_ent = TopAllocArea;
  TopAllocArea += TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_tab_ent;
  for (i = 1; i < TABLE_ENTRIES_ALLOC_NUMBER; i++) {
    TabEnt_next(aux) = (tab_ent_ptr) ((unsigned int)aux + TABLE_ENTRY_SIZE);
    aux = TabEnt_next(aux);
  }
  TabEnt_next(aux) = NULL;
#ifdef TABLING_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_tab_ent += TABLE_ENTRIES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* TABLING_STATISTICS */ 
  return;
}


void alloc_subgoal_trie_nodes(void) {
  int i;
  sg_node_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d subgoal trie nodes", SUBGOAL_TRIE_NODES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef YAPOR
  if (SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_subgoal_trie_nodes)");
  GLOBAL_next_free_sg_node = TopAllocArea;
  TopAllocArea += SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_sg_node;
  for (i = 1; i < SUBGOAL_TRIE_NODES_ALLOC_NUMBER; i++) {
    TrNode_next(aux) = (sg_node_ptr) ((unsigned int)aux + SUBGOAL_TRIE_NODE_SIZE);
    aux = TrNode_next(aux);
  }
  TrNode_next(aux) = NULL;
#ifdef TABLING_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_sg_node += SUBGOAL_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* TABLING_STATISTICS */ 
  return;
}


void alloc_subgoal_frames(void) {
  int i;
  sg_fr_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d subgoal frames", SUBGOAL_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef YAPOR
  if (SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_subgoal_frames)");
  GLOBAL_next_free_sg_fr = TopAllocArea;
  TopAllocArea += SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_sg_fr;
  for (i = 1; i < SUBGOAL_FRAMES_ALLOC_NUMBER; i++) {
    SgFr_next(aux) = (sg_fr_ptr) ((unsigned int)aux + SUBGOAL_FRAME_SIZE);
    aux = SgFr_next(aux);
  }
  SgFr_next(aux) = NULL;
#ifdef TABLING_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_sg_fr += SUBGOAL_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* TABLING_STATISTICS */ 
  return;
}


void alloc_answer_trie_nodes(void) {
  int i;
  ans_node_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d answer trie nodes", ANSWER_TRIE_NODES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef YAPOR
  if (ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_answer_trie_nodes)");
  GLOBAL_next_free_ans_node = TopAllocArea;
  TopAllocArea += ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_ans_node;
  for (i = 1; i < ANSWER_TRIE_NODES_ALLOC_NUMBER; i++) {
    TrNode_next(aux) = (ans_node_ptr) ((unsigned int)aux + ANSWER_TRIE_NODE_SIZE);
    aux = TrNode_next(aux);
  }
  TrNode_next(aux) = NULL;
#ifdef TABLING_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_ans_node += ANSWER_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* TABLING_STATISTICS */ 
  return;
}


void alloc_dependency_frames(void) {
  int i;
  dep_fr_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d dependency frames", DEPENDENCY_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
#ifdef YAPOR
  if (DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
#else
  if (DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER > TopWorkArea - TopAllocArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_dependency_frames)");
  GLOBAL_next_free_dep_fr = TopAllocArea;
  TopAllocArea += DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_dep_fr;
  for (i = 1; i < DEPENDENCY_FRAMES_ALLOC_NUMBER; i++) {
    DepFr_next(aux) = (dep_fr_ptr) ((unsigned int)aux + DEPENDENCY_FRAME_SIZE);
    aux = DepFr_next(aux);
  }
  DepFr_next(aux) = NULL;
#ifdef TABLING_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_dep_fr += DEPENDENCY_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* TABLING_STATISTICS */ 
  return;
}


#ifdef YAPOR
void alloc_suspension_frames(void) {
  int i;
  susp_fr_ptr aux;

  TABLING_DEBUG_MESSAGE("Allocating %d suspension frames", SUSPENSION_FRAMES_ALLOC_NUMBER);
  LOCK(GLOBAL_LOCKS_alloc_area);
  if (SUSPENSION_FRAME_SIZE * SUSPENSION_FRAMES_ALLOC_NUMBER > SuspensionArea - TopAllocArea)
    abort_optyap("no more free alloc space (alloc_suspension_frames)");
  GLOBAL_next_free_susp_fr = TopAllocArea;
  TopAllocArea += SUSPENSION_FRAME_SIZE * SUSPENSION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_LOCKS_alloc_area);

  aux = GLOBAL_next_free_susp_fr;
  for (i = 1; i < SUSPENSION_FRAMES_ALLOC_NUMBER; i++) {
    SuspFr_next(aux) = (susp_fr_ptr) ((unsigned int)aux + SUSPENSION_FRAME_SIZE);
    aux = SuspFr_next(aux);
  }
  SuspFr_next(aux) = NULL;
#ifdef OPTYAP_STATISTICS
  LOCK(GLOBAL_STATS_lock);
  GLOBAL_STATS_allocated_susp_fr += SUSPENSION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_STATS_lock);
#endif /* OPTYAP_STATISTICS */ 
  return;
}


void *alloc_suspension_block(int size) {
  void *block;
  TABLING_DEBUG_MESSAGE("Allocating suspension block with %d bytes", size);
  LOCK(GLOBAL_LOCKS_alloc_area);
  if (size > SuspensionArea - TopAllocArea)
    abort_optyap("no more free alloc space (alloc_suspension_block)");
  SuspensionArea -= size;
  block = SuspensionArea;
  UNLOCK(GLOBAL_LOCKS_alloc_area);
  return block;
}


void reset_suspension_area(void) {
  SuspensionArea = TopWorkArea;
  return;
}
#endif /* YAPOR */
#endif /* TABLING */
#endif /* YAPOR || TABLING */
