/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		eval.c							 *
* Last rev:								 *
* mods:									 *
* comments:	arithmetical expression evaluation			 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif

/*
 * This file implements arithmetic operations 
 *
 */

#include "Yap.h"
#include "Yatom.h"
#include "Heap.h"
#include "eval.h"

STATIC_PROTO(Int p_is, (void));

yap_error_number YAP_matherror = NO_ERROR;

InitEntry InitTab[] = {
  {"+", 2, e_plus},
  {"-", 2, e_dif},
  {"*", 2, e_times},
  {"/", 2, e_div},
  {"<<", 2, e_lshift},
  {">>", 2, e_rshift},
  {"/\\", 2, e_and},
  {"\\/", 2, e_or},
  {"#", 2, e_xor},
  {"-", 1, e_uminus},
  {"\\", 1, e_not},
  {"exp", 1, e_exp},
  {"log", 1, e_log},
  {"log10", 1, e_log10},
  {"sqrt", 1, e_sqrt},
  {"sin", 1, e_sin},
  {"cos", 1, e_cos},
  {"tan", 1, e_tan},
  {"sinh", 1, e_sinh},
  {"cosh", 1, e_cosh},
  {"tanh", 1, e_tanh},
  {"asin", 1, e_asin},
  {"acos", 1, e_acos},
  {"atan", 1, e_atan},
  {"atan2", 2, e_atan2},
  {"asinh", 1, e_asinh},
  {"acosh", 1, e_acosh},
  {"atanh", 1, e_atanh},
  {"mod", 2, e_mod},
  {"rem", 2, e_mod},
  {"//", 2, e_idiv},
  /* C-Prolog exponentiation */
  {"^", 2, e_power},
  /* ISO-Prolog exponentiation */
  {"**", 2, e_power},
  /* Quintus exponentiation */
  {"exp", 2, e_power},
  {"floor", 1, e_floor},
  {"ceiling", 1, e_ceiling},
  {"round", 1, e_round},
  {"truncate", 1, e_truncate},
  {"+", 1, e_uplus},
  {"integer", 1, e_integer},
  {"float", 1, e_float},
  {"abs", 1, e_abs},
  {"gcd", 2, e_gcd},
  {"min", 2, e_min},
  {"max", 2, e_max},
  {"msb", 1, e_msb},
  {"float_fractional_part", 1, e_fmodf},
  {"float_integer_part", 1, e_imodf},
  {"sign", 1, e_sign}
};

#define NOps sizeof(InitTab)/sizeof(*InitTab)

#ifndef PI
#ifdef M_PI
#define PI M_PI
#else
#define PI 3.14159265358979323846
#endif
#endif

#ifndef M_E
#define M_E 2.7182818284590452354
#endif

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif

#ifndef NAN
#define NAN      (0.0/0.0)
#endif


Int gcd(Int m11,Int m21)
{
  /* Blankinship algorithm, provided by Miguel Filgueiras */
  Int m12=1, m22=0, k;

  while (m11>0 && m21>0)
    if (m11<m21) {
      k = m21/m11;  m21 -= k*m11;  m22 -= k*m12;
    } else {
      k=m11/m21;  m11 -= k*m21;  m12 -= k*m22;
    }
  if (m11<0 || m21<0) {		/* overflow? */
/*    Oflow = 1; */
    Error(EVALUATION_ERROR_INT_OVERFLOW, MkIntegerTerm(m11),
	  "gcd/2 with %d and %d", m11, m21);
    P = (yamop *)FAILCODE;
    return(1);
  }
  if (m11)  return(m11);
  return(m21);
}

Int gcdmult(Int m11,Int m21,Int *pm11)	/* *pm11 gets multiplier of m11 */
{
  Int m12=1, m22=0, k;

  while (m11 && m21)
    if (m11<m21) {
      k = m21/m11;  m21 -= k*m11;  m22 -= k*m12;
    } else {
      k=m11/m21;  m11 -= k*m21;  m12 -= k*m22;
    }
  if (m11<0 || m21<0) {		/* overflow? */
/*    Oflow = 1; */
    Error(EVALUATION_ERROR_INT_OVERFLOW, MkIntegerTerm(m11),
	  "gcdmult/2 with %d and %d", m11, m21);
    P = (yamop *)FAILCODE;
    return(1);
  }
  if (m11) {
    *pm11 = m12;  return(m11);
  }
  *pm11 = m22;
  return(m21);
}

Int msb(Int inp)	/* calculate the most significant bit for an integer */
{
  /* the obvious solution: do it by using binary search */
  Int out = 0;
  int off = sizeof(CELL)*4;

  if (inp < 0) {
    Error(DOMAIN_ERROR_NOT_LESS_THAN_ZERO, MkIntegerTerm(inp),
	  "msb/1 received %d", inp);
    P = (yamop *)FAILCODE;
    return(0);
  }

  while (off) {
    int limit = 1 << (off);
    if (inp >= limit) {
      out += off;
      inp >>= off;
    }
    off >>= 1;
  }
  return(out);
}

void
InitEval(void)
{
  unsigned int    i;
  ExpEntry       *p;
  Atom            a;

  for (i = 0; i < NOps; ++i) {
    a = LookupAtom(InitTab[i].OpName);
    if (GetExpProp(a, InitTab[i].arity))
      break;
    p = (ExpEntry *) AllocAtomSpace(sizeof(ExpEntry));
    p->KindOfPE = ExpProperty;
    p->ArityOfEE = InitTab[i].arity;
    p->ENoOfEE = InitTab[i].eno;
    p->NextOfPE = RepAtom(a)->PropOfAE;
    RepAtom(a)->PropOfAE = AbsExpProp(p);
  }
  /* here are the arithmetical predicates */
  InitCPred("is", 2, p_is, TestPredFlag | SafePredFlag);
  InitCPred("is", 3, p_unary_is, TestPredFlag | SafePredFlag);
  InitCPred("is", 4, p_binary_is, TestPredFlag | SafePredFlag);
}

Term current_eval_term;

/*
  try to find out why we had an error,
   as we cannot always obtain reliable information from the system 
*/
yap_error_number
EvalError(Term rt) {
  int function;
  Term a1, a2;

  if (rt == TermNil) {
    /* in this case we do not know the registers so we cannot do much.
       Let's just guess it is an overflow because I don't know how
       it can be anything else.  */
    return(EVALUATION_ERROR_FLOAT_OVERFLOW);
  } else if (IsIntTerm(rt)) {
    Int i = IntOfTerm(rt);
    Term t2 = Deref(ARG2);
    if (IsAtomTerm(t2)) {
      Atom            name;
      Prop            p;

      name = AtomOfTerm(t2);
      if ((p = GetExpProp(name, 1)) == NIL) {
	return(EVALUATION_ERROR_UNDEFINED);
      }
      function = RepExpProp(p)->ENoOfEE;
    } else if (IsIntTerm(t2))
      function = IntOfTerm(t2);
    else
    return(EVALUATION_ERROR_UNDEFINED);
    if (i == 1) {
      a1 = Deref(ARG3);
      a2 = Deref(TermNil);
    } else {
      a1 = Deref(ARG3);
      a2 = Deref(ARG4);
    }
  } else if (IsApplTerm(rt)) {
    UInt            n;
    Functor	    fun = FunctorOfTerm(rt);
    Atom            name;
    register Prop   p;

    n = ArityOfFunctor(fun);
    name  = NameOfFunctor(fun);
    if ((p = GetExpProp(name, n)) == NIL) {
      /* should never happen */
      return(TYPE_ERROR_EVALUABLE);
    }
    function = RepExpProp(p)->ENoOfEE;
    a1 = Deref(ArgOfTerm(1,rt));
    if (n > 1)
      a2 = Deref(ArgOfTerm(2,rt));
    else
      a2 = TermNil; /* just make gcc happy */
  } else {
    return(EVALUATION_ERROR_UNDEFINED);
  }
  switch (function) {
  case e_div:
  case e_mod:
  case e_idiv:
    if (a2 == MkIntTerm(0) ||
	FloatOfTerm(a2) == FL(0.0)) {
      return(EVALUATION_ERROR_ZERO_DIVISOR);
    }
  case e_plus:
  case e_dif:
  case e_times:
  case e_xor:
  case e_or:
  case e_lshift:
  case e_rshift:
  case e_power:
  case e_min:
  case e_max:
  case e_gcd:
  case e_gcdmult:
    if (IsIntegerTerm(a1) && IsIntegerTerm(a2))
      return(EVALUATION_ERROR_INT_OVERFLOW);
    else
      return(EVALUATION_ERROR_FLOAT_OVERFLOW);
  case e_abs:
  case e_uminus:
  case e_uplus:
  case e_msb:
  case e_not:
  case e_exp:
  case e_sin:
  case e_cos:
  case e_tan:
  case e_sinh:
  case e_cosh:
  case e_tanh:
  case e_sqrt:
  case e_atan:
  case e_asinh:
  case e_acosh:
  case e_atanh:
  case e_atan2:
  case e_floor:
  case e_round:
  case e_ceiling:
  case e_truncate:
  case e_integer:
  case e_float:
  case e_fmodf:
  case e_imodf:
  case e_sign:
    if (IsIntegerTerm(a1))
      return(EVALUATION_ERROR_INT_OVERFLOW);
    else
      return(EVALUATION_ERROR_FLOAT_OVERFLOW);
  case e_log:
  case e_log10:
    {
      if (IsFloatTerm(a1) && FloatOfTerm(a1) < 0.0) {
	return(EVALUATION_ERROR_UNDEFINED);
      }
    }
    if (IsIntegerTerm(a1))
      return(EVALUATION_ERROR_INT_OVERFLOW);
    else
      return(EVALUATION_ERROR_FLOAT_OVERFLOW);
  case e_asin:
  case e_acos:
    {
      if (IsFloatTerm(a1) &&
	  (FloatOfTerm(a1) < 1.0 || FloatOfTerm(a1) < 1.0)) {
	return(EVALUATION_ERROR_UNDEFINED);
      }
    }
    if (IsIntegerTerm(a1))
      return(EVALUATION_ERROR_INT_OVERFLOW);
    else
      return(EVALUATION_ERROR_FLOAT_OVERFLOW);
  default:
    return(EVALUATION_ERROR_UNDEFINED);
  }
}

int 
UnEvalFl(BITS16 op, Float fl1)
{
  switch(op) {
  case e_uminus:
    REvalFl(-fl1);
  case e_abs:
    REvalFl(fabs(fl1));
  case e_uplus:
    REvalFl(fl1);
  case e_exp:
    REvalFl(exp(fl1));
  case e_log:
    REvalFl(log(fl1));
  case e_log10:
    REvalFl(log10(fl1));
  case e_sqrt:
    REvalFl(sqrt(fl1));
  case e_sin:
    REvalFl(sin(fl1));
  case e_cos:
    REvalFl(cos(fl1));
  case e_tan:
    REvalFl(tan(fl1));
  case e_sinh:
    REvalFl(sinh(fl1));
  case e_cosh:
    REvalFl(cosh(fl1));
  case e_tanh:
    REvalFl(tanh(fl1));
  case e_asin:
    REvalFl(asin(fl1));
  case e_acos:
    REvalFl(acos(fl1));
  case e_atan:
    REvalFl(atan(fl1));
  case e_asinh:
    REvalFl(asinh(fl1));
  case e_acosh:
    REvalFl(acosh(fl1));
  case e_atanh:
    REvalFl(atanh(fl1));
  case e_floor:
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) { /* iso */
      Float out = floor(fl1);
      if (out <= (Float)Int_MAX && out >= (Float)Int_MIN) {
	REvalInt((Int) out);
      } else {
	Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fl1), "round/1");
	P = (yamop *)FAILCODE;
	REvalError();	    
      }
    } else {
      REvalFl(floor(fl1));
    }
  case e_ceiling:
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) { /* iso */
      Float out = ceil(fl1);
      if (out <= (Float)Int_MAX && out >= (Float)Int_MIN) {
	REvalInt((Int) out);
      } else {
	Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fl1), "round/1");
	P = (yamop *)FAILCODE;
	REvalError();	    
      }
    } else {
      REvalFl(ceil(fl1));
    }
  case e_round:
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) { /* iso */
      Float d = fl1 - floor(fl1), out;
      /* let us implement our own rounding function */
      if (d >= 0.5) {
	out = ceil(fl1);
      } else {
	out = floor(fl1);
      }
      if (out <= (Float)Int_MAX && out >= (Float)Int_MIN) {
	REvalInt((Int) out);
      } else {
	Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fl1), "round/1");
	P = (yamop *)FAILCODE;
	REvalError();	    
      }
    } else {
      REvalFl(rint(fl1));
    }
  case e_truncate:
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) { /* iso */
      Float out = (fl1 >= 0.0 ? floor(fl1) : ceil(fl1));
      if (out <= (Float)Int_MAX && out >= (Float)Int_MIN) {
	REvalInt((Int) out);
      } else {
	Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fl1), "round/1");
	P = (yamop *)FAILCODE;
	REvalError();	    
      }
    } else {
      REvalFl((fl1 >= 0.0 ? floor(fl1) : ceil(fl1)) );
    }
  case e_integer:
    if (fl1 <= (Float)Int_MAX && fl1 >= (Float)Int_MIN) {
      REvalInt((Int) fl1);
    } else {
      Error(EVALUATION_ERROR_INT_OVERFLOW, MkFloatTerm(fl1), "integer/1");
      P = (yamop *)FAILCODE;
      REvalError();	    
    }
  case e_float:
    REvalFl(fl1);
  case e_fmodf:
    {
      Float dbl;
      REvalFl(modf(fl1,&dbl));
    }
  case e_imodf:
    {
      Float dbl;
      
      modf(fl1,&dbl);
      REvalFl(dbl);
    }
  case e_sign:
    if (fl1 < 0) {
      REvalFl(FL(-1.0));
    } else if (fl1 == (Float)0.0) {
      REvalFl(FL(0.0));
    } else {
      REvalFl(FL(1.0));
    }
  case e_not:
    Error(TYPE_ERROR_INTEGER, MkFloatTerm(fl1), "not/1");
    P = (yamop *)FAILCODE;
    REvalError();
  case e_msb:
    Error(TYPE_ERROR_INTEGER, MkFloatTerm(fl1), "msb/1");
    P = (yamop *)FAILCODE;
    REvalError();
  default:
    {
      Term t, ti[2];
 
      ti[0] = MkAtomTerm(NameOfFunctor(FunctorOfTerm(current_eval_term)));
      ti[1] = MkIntegerTerm(1);
      t = MkApplTerm(MkFunctor(LookupAtom("/"),1), 1, ti);
      Error(TYPE_ERROR_EVALUABLE, t,
	    "arithmetic expression  %s/%d",
	    RepAtom(NameOfFunctor(FunctorOfTerm(current_eval_term)))->StrOfAE,
	    2
	    );
      P = (yamop *)FAILCODE;
      REvalError();
    }
  }
}

int
Eval(Term t)
{
  register Term   rt = Deref(t);
  
  current_eval_term = rt;
  if (IsVarTerm(rt)) {
    Error(INSTANTIATION_ERROR, rt, "arithmetic expression");
    P = (yamop *)FAILCODE;
    REvalError();
  }
  if (IsIntTerm(rt))
    REvalInt(IntOfTerm(rt));
  if (IsApplTerm(rt)) {
    Int             n;
    Functor	    fun = FunctorOfTerm(rt);
    Atom            name;
    register Prop   p;

    if (IsExtensionFunctor(fun)) {
      if (fun == FunctorDouble) {
	REvalFl(FloatOfTerm(rt));
      } else if (fun == FunctorLongInt) {
	REvalInt(LongIntOfTerm(rt));
      } else {
	Error(TYPE_ERROR_EVALUABLE, rt, "arithmetic expression");
	P = (yamop *)FAILCODE;
        REvalError();
      }
    }
    n = ArityOfFunctor(fun);
    name  = NameOfFunctor(fun);
    if ((p = GetExpProp(name, n)) == NIL) {
      Term t, ti[2];

      ti[0] = MkAtomTerm(NameOfFunctor(FunctorOfTerm(current_eval_term)));
      ti[1] = MkIntegerTerm(n);
      t = MkApplTerm(MkFunctor(LookupAtom("/"),2), 2, ti);
      Error(TYPE_ERROR_EVALUABLE, t,
	    "functor %s/%d for arithmetic expression",
	    RepAtom(NameOfFunctor(FunctorOfTerm(current_eval_term)))->StrOfAE,
	    n
	    );
      P = (yamop *)FAILCODE;
      REvalError();
    }
    if (n > 1) {
      Term t0 = ArgOfTerm(1, rt);
      Term t1 = ArgOfTerm(2, rt);
      Int I0;
      Float F0;
      
      if (IsVarTerm(t0)) {
	Error(INSTANTIATION_ERROR, t,
	      "functor %s/%d for arithmetic expression",
	      RepAtom(NameOfFunctor(FunctorOfTerm(current_eval_term)))->StrOfAE,
	      n
	      );
	P = (yamop *)FAILCODE;
	REvalError();
      }
      if (IsIntegerTerm(t0)) {
	Float F1;
	Int I1;

	I0 = IntegerOfTerm(t0);
      t0_int:
	if (IsVarTerm(t1)) {
	  Error(INSTANTIATION_ERROR, t,
		"functor %s/%d for arithmetic expression",
		RepAtom(NameOfFunctor(FunctorOfTerm(current_eval_term)))->StrOfAE,
		n
		);
	  P = (yamop *)FAILCODE;
	  REvalError();
	}
	if (IsIntegerTerm(t1)) {
	  I1 = IntegerOfTerm(t1);
	t0_int_t1_int:
	  current_eval_term = rt;
	  return(BinEvalInt(RepExpProp(p)->ENoOfEE, I0, I1));
	} else if (IsFloatTerm(t1)) {
	  F1 = FloatOfTerm(t1);
	t0_int_t1_flt:
	  current_eval_term = rt;
	  return(BinEvalFl(RepExpProp(p)->ENoOfEE, FL(I0), F1, 2));
	} else {
	  int flag = Eval(t1);
	  if (flag == FError) {
	    return(FError);
	  } else if (flag == FInt) {
	    I1 = eval_int;
	    goto t0_int_t1_int;
	  } else {
	    F1 = eval_flt;
	    goto t0_int_t1_flt;
	  }
	}
      } else if (IsFloatTerm(t0)) {
	Float F1;
	Int I1;

	F0 = FloatOfTerm(t0);
      t0_flt:
	if (IsIntegerTerm(t1)) {
	  I1 = IntegerOfTerm(t1);
	t0_flt_t1_int:
	  current_eval_term = rt;
	  return(BinEvalFl(RepExpProp(p)->ENoOfEE, F0, FL(I1), 1));
	} else if(IsFloatTerm(t1)) {
	  F1 = FloatOfTerm(t1);
	t0_flt_t1_flt:
	  current_eval_term = rt;
	  return(BinEvalFl(RepExpProp(p)->ENoOfEE, F0, F1, 3));
	} else {
	  int flag = Eval(t1);
	  if (flag == FError) {
	    return(FError);
	  } else if (flag == FInt) {
	    I1 = eval_int;
	    goto t0_flt_t1_int;
	  } else {
	    F1 = eval_flt;
	    goto t0_flt_t1_flt;
	  }
	}
      } else {
	int Fgop1 = Eval(t0);
	if (Fgop1 == FError) {
	  return(FError);
	} else if (Fgop1 == FInt) {
	  I0 = eval_int;
	  goto t0_int;
	} else {
	  F0 = eval_flt;
	  goto t0_flt;
	} 
      }
    } else { /* n == 1 */
      Term            t0 = ArgOfTerm(1, rt);
	
      if (IsVarTerm(t0)) {
	Error(INSTANTIATION_ERROR, t,
	      "functor %s/%d for arithmetic expression",
	      RepAtom(NameOfFunctor(FunctorOfTerm(current_eval_term)))->StrOfAE,
	      n
	      );
	P = (yamop *)FAILCODE;
	REvalError();
      }
      if (IsIntegerTerm(t0)) {
	return(UnEvalInt(RepExpProp(p)->ENoOfEE,IntegerOfTerm(t0)));
      } else if (IsFloatTerm(t0)) {
	return(UnEvalFl(RepExpProp(p)->ENoOfEE,FloatOfTerm(t0)));
      } else {
	int Fgop = Eval(t0);
	
	current_eval_term = rt;
	if (Fgop == FError)
	  return(FError);
	else if (Fgop == FInt)
	  return(UnEvalInt(RepExpProp(p)->ENoOfEE,eval_int));
	else
	  return(UnEvalFl(RepExpProp(p)->ENoOfEE,eval_flt));
      }
    }
  }
  if (IsAtomTerm(rt)) {
    register Atom   art = AtomOfTerm(rt);
    if (art == AtomPi) {
      REvalFl(PI);
    } else if (art == AtomE) {
      REvalFl(M_E);
#ifndef _MSC_VER
    } else if (art == AtomInf) {
      if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {/* iso */
	Error(TYPE_ERROR_EVALUABLE, rt, "evaluating infinity");
	P = (yamop *)FAILCODE;
	REvalError();
      } else {
	REvalFl(INFINITY);
      }
    } else if (art == AtomNan) {
      if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {/* iso */
	Error(TYPE_ERROR_EVALUABLE, rt, "evaluating not-a-number");
	P = (yamop *)FAILCODE;
	REvalError();
      } else {
	REvalFl(NAN);
      }
#endif
    } else if (art == AtomRandom) {
      REvalFl(yap_random());
    } else if (art == AtomCpuTime) {
      
      REvalFl((Float)cputime()/1000.0);
    } if (art == AtomHeapUsed) {
      REvalInt(HeapUsed);
    /*
     * # i f d e f 
     *
     * LO N G _ A D D R E S S E S 
     *
     */
    } else if (art == AtomLocal) {
#if SBA
      REvalInt((Int)ASP);
#else
      REvalInt(LCL0 - ASP);
#endif
    } else if (art == AtomB) {
#if SBA
      REvalInt((Int)B);
#else
      REvalInt(LCL0 - (CELL *)B);
#endif
    } else if (art == AtomGlobal) {
#if SBA
      REvalInt((Int)H);
#else
      REvalInt(H - H0);
#endif
    } else if (art == AtomStackFree) {
      REvalInt(Unsigned(ASP) - Unsigned(H));
    } else {
      Term tar[2];
      tar[0] = rt;
      tar[1] = MkIntTerm(0);
      Error(TYPE_ERROR_EVALUABLE, rt, "evaluating atom %s", RepAtom(art)->StrOfAE);
    }
    P = (yamop *)FAILCODE;
    REvalError();
  }
  /* must be a list */
  return (Eval(HeadOfTerm(rt)));
}



Term 
p_succ(Term TEntry)
{
  if (IsVarTerm(TEntry)) {
    Error(INSTANTIATION_ERROR, TEntry, "+/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsFloatTerm(TEntry)) {
    register Float  f = FloatOfTerm(TEntry);
    return (MkFloatTerm(++f));
  }
  else if (IsLongIntTerm(TEntry)) {
    register Int  i = LongIntOfTerm(TEntry);
    return (MkIntegerTerm(++i));
  }
  if (FInt == Eval(TEntry))
    return (MkIntegerTerm(++(eval_int)));
  return (MkFloatTerm(++(eval_flt)));
}

Term 
p_predc(Term TEntry)
{

  if (IsVarTerm(TEntry)) {
    Error(INSTANTIATION_ERROR, TEntry, "+/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsFloatTerm(TEntry)) {
    register Float  f = FloatOfTerm(TEntry);
    return (MkFloatTerm(--f));
  }
  else if (IsLongIntTerm(TEntry)) {
    register Int  i = LongIntOfTerm(TEntry);
    return (MkIntegerTerm(--i));
  }
  if (FInt == Eval(ARG1))
    return (MkIntegerTerm(--eval_int));
  return (MkFloatTerm(--eval_flt));
}



Term 
p_plus(Term TFirst, Term TSec)
{
  if (IsVarTerm(TFirst) || IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, (IsVarTerm(TFirst) ? TFirst : TSec) , "+/2");
    P = (yamop *)FAILCODE;
    return(MkIntTerm(0));
  }
  if (IsIntegerTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkIntegerTerm(IntegerOfTerm(TFirst)+IntegerOfTerm(TSec)));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FL(IntegerOfTerm(TFirst))+FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkIntegerTerm(IntegerOfTerm(TFirst)+eval_int));
      } else {
	return(MkFloatTerm(FL(IntegerOfTerm(TFirst))+eval_flt));
      }
    }
  } else if (IsFloatTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)+FL(IntegerOfTerm(TSec))));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)+FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkFloatTerm(FloatOfTerm(TFirst)+FL(eval_int)));
      } else {
	return(MkFloatTerm(FloatOfTerm(TFirst)+eval_flt));
      }
    }
  } else {
    int aflag = Eval(TFirst);
    if (aflag == FError) {
      P = (yamop *)FAILCODE;
      return(MkIntTerm(0));
    } else if (aflag == FInt) {
      return(p_plus(MkIntegerTerm(eval_int),TSec));
    } else {
      return(p_plus(MkFloatTerm(eval_flt),TSec));
    }
  }
}

Term 
p_minus(Term TFirst, Term TSec)
{
  if (IsVarTerm(TFirst) || IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, (IsVarTerm(TFirst) ? TFirst : TSec) , "-/2");
    P = (yamop *)FAILCODE;
    return(MkIntTerm(0));
  }
  if (IsIntegerTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkIntegerTerm(IntegerOfTerm(TFirst)-IntegerOfTerm(TSec)));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FL(IntegerOfTerm(TFirst))-FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkIntegerTerm(IntegerOfTerm(TFirst)-eval_int));
      } else {
	return(MkFloatTerm(FL(IntegerOfTerm(TFirst))-eval_flt));
      }
    }
  } else if (IsFloatTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)-FL(IntegerOfTerm(TSec))));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)-FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkFloatTerm(FloatOfTerm(TFirst)-FL(eval_int)));
      } else {
	return(MkFloatTerm(FloatOfTerm(TFirst)-eval_flt));
      }
    }
  } else {
    int aflag = Eval(TFirst);
    if (aflag == FError) {
      P = (yamop *)FAILCODE;
      return(MkIntTerm(0));
    } else if (aflag == FInt) {
      return(p_minus(MkIntegerTerm(eval_int),TSec));
    } else {
      return(p_minus(MkFloatTerm(eval_flt),TSec));
    }
  }
}

Term 
p_times(Term TFirst, Term TSec)
{
  if (IsVarTerm(TFirst) || IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, (IsVarTerm(TFirst) ? TFirst : TSec) , "*/2");
    P = (yamop *)FAILCODE;
    return(MkIntTerm(0));
  }
  if (IsIntegerTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkIntegerTerm(IntegerOfTerm(TFirst)*IntegerOfTerm(TSec)));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FL(IntegerOfTerm(TFirst))*FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkIntegerTerm(IntegerOfTerm(TFirst)*eval_int));
      } else {
	return(MkFloatTerm(FL(IntegerOfTerm(TFirst))*eval_flt));
      }
    }
  } else if (IsFloatTerm(TFirst)) {
    if (IsIntegerTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)*FL(IntegerOfTerm(TSec))));
    } else if (IsFloatTerm(TSec)) {
      return(MkFloatTerm(FloatOfTerm(TFirst)*FloatOfTerm(TSec)));
    } else {
      int aflag = Eval(TSec);
      if (aflag == FError) {
	P = (yamop *)FAILCODE;
	return(MkIntTerm(0));
      } else if (aflag == FInt) {
	return(MkFloatTerm(FloatOfTerm(TFirst)*FL(eval_int)));
      } else {
	return(MkFloatTerm(FloatOfTerm(TFirst)*eval_flt));
      }
    }
  } else {
    int aflag = Eval(TFirst);
    if (aflag == FError) {
      P = (yamop *)FAILCODE;
      return(MkIntTerm(0));
    } else if (aflag == FInt) {
      return(p_times(MkIntegerTerm(eval_int),TSec));
    } else {
      return(p_times(MkFloatTerm(eval_flt),TSec));
    }
  }
}

Term 
p_div(Term TFirst, Term TSec)
{
  Int r1, r2;

  if (IsVarTerm(TFirst)) {
    Error(INSTANTIATION_ERROR, TFirst, "/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, TSec, "/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsIntTerm(TFirst)) {
    if (IsIntTerm(TSec))
      return (MkIntegerTerm(IntOfTerm(TFirst) /
			IntOfTerm(TSec)));
    else
      r1 = IntOfTerm(TFirst);
  } else if (IsFloatTerm(TFirst) || (Eval(TFirst) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TFirst, "// /2");
    /* make GCC happy */
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r1 = eval_int;
  if (IsIntTerm(TSec))
    r2 = IntOfTerm(TSec);
  else  if (IsFloatTerm(TSec) || (Eval(TSec) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TSec, "// /2");
    /* make GCC happy */
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r2 = eval_int;
  current_eval_term = TermNil;
  if (r2 == 0) {
    Error(EVALUATION_ERROR_ZERO_DIVISOR, TSec, "// /2");
    /* make GCC happy */
    P = (yamop *)FAILCODE;
    return(0L);
  }
  return (MkIntegerTerm(r1 / r2));
}

Term 
p_and(Term TFirst, Term TSec)
{
  Int r1, r2;

  if (IsVarTerm(TFirst)) {
    Error(INSTANTIATION_ERROR, TFirst, "/\\/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, TSec, "/\\/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsIntTerm(TFirst)) {
    if (IsIntTerm(TSec))
      return (MkIntegerTerm(IntOfTerm(TFirst) &
			IntOfTerm(TSec)));
    else
      r1 = IntOfTerm(TFirst);
  } else if (IsFloatTerm(TFirst) || (Eval(TFirst) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TFirst, "/\\/2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r1 = eval_int;
  if (IsIntTerm(TSec))
    r2 = IntOfTerm(TSec);
  else  if (IsFloatTerm(TSec) || (Eval(TSec) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TSec, "/\\/2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r2 = eval_int;
  current_eval_term = TermNil;
  return (MkIntegerTerm(r1 & r2));
}

Term 
p_or(Term TFirst, Term TSec)
{
  Int r1, r2;

  if (IsVarTerm(TFirst)) {
    Error(INSTANTIATION_ERROR, TFirst, "\\/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, TSec, "\\/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsIntTerm(TFirst)) {
    if (IsIntTerm(TSec))
      return (MkIntegerTerm(IntOfTerm(TFirst) |
			IntOfTerm(TSec)));
    else
      r1 = IntOfTerm(TFirst);
  } else if (IsFloatTerm(TFirst) || (Eval(TFirst) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TFirst, "\\/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r1 = eval_int;
  if (IsIntegerTerm(TSec))
    r2 = IntegerOfTerm(TSec);
  else  if (IsFloatTerm(TSec) || (Eval(TSec) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TSec, "\\/ /2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r2 = eval_int;
  current_eval_term = TermNil;
  return (MkIntegerTerm(r1 | r2));
}

Term 
p_sll(Term TFirst, Term TSec)
{
  Int r1, r2;

  if (IsVarTerm(TFirst)) {
    Error(INSTANTIATION_ERROR, TFirst, "<</2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, TSec, "<</2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsIntTerm(TFirst)) {
    if (IsIntTerm(TSec))
      return (MkIntegerTerm(IntOfTerm(TFirst) <<
			IntOfTerm(TSec)));
    else
      r1 = IntOfTerm(TFirst);
  } else if (IsFloatTerm(TFirst) || (Eval(TFirst) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TFirst, "<</2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r1 = eval_int;
  if (IsIntegerTerm(TSec))
    r2 = IntegerOfTerm(TSec);
  else  if (IsFloatTerm(TSec) || (Eval(TSec) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TSec, "<</2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r2 = eval_int;
  current_eval_term = TermNil;
  return (MkIntegerTerm(r1 << r2));
}

Term 
p_slr(Term TFirst, Term TSec)
{
  Int r1, r2;

  if (IsVarTerm(TFirst)) {
    Error(INSTANTIATION_ERROR, TFirst, ">>/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsVarTerm(TSec)) {
    Error(INSTANTIATION_ERROR, TSec, ">>/2");
    P = (yamop *)FAILCODE;
    return(0L);
  }
  if (IsIntegerTerm(TFirst)) {
    if (IsIntegerTerm(TSec))
      return (MkIntegerTerm(IntegerOfTerm(TFirst) >>
			IntegerOfTerm(TSec)));
    else
      r1 = IntOfTerm(TFirst);
  } else if (IsFloatTerm(TFirst) || (Eval(TFirst) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TFirst, ">>/2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r1 = eval_int;
  if (IsIntegerTerm(TSec))
    r2 = IntegerOfTerm(TSec);
  else  if (IsFloatTerm(TSec) || (Eval(TSec) == FFloat)) {
    Error(TYPE_ERROR_INTEGER, TSec, ">>/2");
    P = (yamop *)FAILCODE;
    return(0L);
  } else
    r2 = eval_int;
  current_eval_term = TermNil;
  return (MkIntegerTerm(r1 >> r2));
}

Int 
EvFArt(Term te)
{
  Float           fl;
  int             res;

  current_eval_term = te;
  if (IsFloatTerm(te)) {
    fl = FloatOfTerm(te);
    if (FlIsInt(fl))
      return ((Int) fl);
    else {
      Error(TYPE_ERROR_INTEGER, te, "evaluating functor arity");
      /* make GCC happy */
      res = FFloat;
      P = (yamop *)FAILCODE;
      return(0);
    }
  } else
    res = Eval(te);
  if (res == FInt)
    return (eval_int);
  else if (FlIsInt(eval_flt))
    return ((Int) eval_flt);
  else {
    res = FFloat;
    Error(TYPE_ERROR_INTEGER, te, "evaluating functor arity");
    P = (yamop *)FAILCODE;
  }
  return (0L);		/* to avoid warning */
}

static Int 
p_is(void)
{				/* X is Y	 */
  Term            t;
  int res = Eval(ARG2);
  if (res ==  FInt)
    t = MkIntegerTerm(eval_int);
  else if (res == FFloat) {
    t = MkFloatTerm(eval_flt);
  }  else
    return(FALSE);
  return (unify_constant(ARG1, t));
}

