/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		corout.pl						 *
* Last rev:								 *
* mods:									 *
* comments:	Coroutines implementation				 *
*									 *
*************************************************************************/


%:- module(coroutining,[
%dif/2,
%when/2,
%block/1,
%wait/1,
%frozen/2
%]). 

%
% operators defined in this module:
%
:- op(1150, fx, block).

%
% Tell the system how to present frozen goals.
%
:- assert((extensions_to_present_answer(Level) :-
	   '$show_frozen_goals'(Level))).

/* keep the original goal around in case the frozen variables were there */
/* should always fail */
'$show_list_of_frozen_goals'(LGs0) :- 
	'$sort'(LGs0, LGs),
	'$purge_done_goals'(LGs, LG),  % from corout.yap
	'$convert_frozen_goals'(LG, NLG),
	NLG = [_|_],
	nl(user_error),
	'$present_frozen_goals'(NLG),
	fail.

'$convert_frozen_goals'([], []).
'$convert_frozen_goals'([G|LG], O) :-
	'$convert_frozen_goal'(G, NG),
	( NG = true -> O = ILG ; O = [NG|ILG]),
	'$convert_frozen_goals'(LG, ILG).

'$present_frozen_goals'([H|LGs]) :-
	write(user_error, H),
	'$continue_frozen_goals'(LGs).

'$continue_frozen_goals'([]) :- !.
'$continue_frozen_goals'(LGs) :-
	write(user_error, ','),
	nl(user_error),
	'$present_frozen_goals'(LGs).

%
% wake_up_goal is called by the system whenever a suspended goal
% resumes.
%

/* The first case may happen if this variable was used for dif.
   In this case, we need a way to keep the original
   suspended goal around
*/
%'$wake_up_goal'([Module1|Continuation],G) :-
%	'$write'(4,vsc_woke:G+[Module1|Continuation]:'
%'), fail.
'$wake_up_goal'([Module1|Continuation], LG) :-
	'$execute_woken_system_goals'(LG),
	'$execute'(Module1:Continuation).


'$execute_woken_system_goals'([]). 
'$execute_woken_system_goals'([G|LG]) :-
	'$execute_woken_system_goal'(G, G),
	'$execute_woken_system_goals'(LG).

%
% X and Y may not be bound (multiple suspensions on the same goal).
%
'$execute_woken_system_goal'('$redo_dif'(Done, X, Y), G) :-
	'$redo_dif'(Done, X, Y, G).
%
% X surely was bound, otherwise we would not be awaken.
%
'$execute_woken_system_goal'('$redo_freeze'(Done, _, Goal), _) :-
	'$redo_freeze'(Done, Goal).
'$execute_woken_system_goal'('$redo_eq'(Done, X, Y, Goal), G) :-
	'$redo_eq'(Done, X, Y, Goal, G).
'$execute_woken_system_goal'('$redo_ground'(Done, X, Goal), _) :-
	'$redo_ground'(Done, X, Goal).
'$execute_woken_system_goal'('$att_do'(V,New), _) :-
	attributes:woken_att_do(V,New).

freeze(V, G) :-
	var(V), !,
	'$freeze_goal'(V,G).
freeze(_, G) :-
	'$execute'(G).

'$freeze_goal'(V,VG) :-
	var(VG), !,
	'$current_module'(M,M),
	'$freeze'(V, '$redo_freeze'(_Done,V,M:G)).
'$freeze_goal'(V,M:G) :- !,
	'$freeze'(V, '$redo_freeze'(_Done,V,M:G)).
'$freeze_goal'(V,G) :-
	'$current_module'(M,M),
	'$freeze'(V, '$redo_freeze'(_Done,V,M:G)).

%
%
% Dif is tricky because we need to wake up on the two variables being
% bound together, or on any variable of the term being bound to
% another. Also, the day YAP fully supports infinite rational trees,
% dif should work for them too. Hence, term comparison should not be
% implemented in Prolog.
%
% This is the way dif works. The '$can_unify' predicate does not know
% anything about dif semantics, it just compares two terms for
% equaility and is based on compare. If it succeeds without generating
% a list of variables, the terms are equal and dif fails. If it fails,
% dif succeeds.
%
% If it succeeds but it creates a list of variables, dif creates
% suspension records for all these variables on the '$redo_dif'(V,
% X, Y) goal. V is a flag that says whether dif has completed or not,
% X and Y are the original goals. Whenever one of these variables is
% bound, it calls '$redo_dif' again. '$redo_dif' will then check whether V
% was bound. If it was, dif has succeeded and redo_dif just
% exits. Otherwise, '$redo_dif' will call dif again to see what happened.
%
% Dif needs two extensions from the suspension engine:
%
% First, it needs
% for the engine to be careful when binding two suspended
% variables. Basically, in this case the engine must be sure to wake
% up one of the goals, as they may make dif fail. The way the engine
% does so is by searching the list of suspended variables, and search
% whether they share a common suspended goal. If they do, that
% suspended goal is added to the WokenList.
%
% Second, thanks to dif we may try to suspend on the same variable
% several times. dif calls a special version of freeze that checks
% whether that is in fact the case.
%
dif(X, Y) :- '$can_unify'(X, Y, LVars), !,
	LVars = [_|_],
	'$dif_suspend_on_lvars'(LVars, '$redo_dif'(Done, X, Y)).
dif(_, _).


'$dif_suspend_on_lvars'([], _).
'$dif_suspend_on_lvars'([H|T], G) :-
	'$freeze'(H, G),
	'$dif_suspend_on_lvars'(T, G).

%
% This predicate is called whenever a variable dif was suspended on is
% bound. Note that dif may have already executed successfully.
%
% Three possible cases: dif has executed and Done is bound; we redo
% dif and the two terms either unify, hence we fail, or may unify, and
% we try to increase the number of suspensions; last, the two terms
% did not unify, we are done, so we succeed and bind the Done variable.
%
'$redo_dif'(Done, X, Y, G) :- nonvar(Done), !.
'$redo_dif'(Done, X, Y, G) :-
	'$can_unify'(X, Y, LVars), !,
	LVars = [_|_],
	'$dif_suspend_on_lvars'(LVars, G).
'$redo_dif'('$done', _, _, _).

% If you called nonvar as condition for when, then you may find yourself
% here.
%
% someone else (that is Cond had ;) did the work, do nothing
%
'$redo_freeze'(Done, Goal) :- nonvar(Done), !.
%
% We still have some more conditions: continue the analysis.
%
'$redo_freeze'(Done, '$when'(C, G, Done)) :- !,
	'$when'(C, G, Done).
	
%
% I can't believe it: we're done and can actually execute our
% goal. Notice we have to say we are done, otherwise someone else in
% the disjunction might decide to wake up the goal themselves.
%
'$redo_freeze'('$done', G) :-
	'$execute'(G).

%
% eq is a combination of dif and freeze
'$redo_eq'(Done, X, Y, Goal, G) :- nonvar(Done), !.
'$redo_eq'(Done, X, Y, _, G) :-
	'$can_unify'(X, Y, LVars),
	LVars = [_|_], !,
	'$dif_suspend_on_lvars'(LVars, G).
'$redo_eq'(Done, _, _, '$when'(C, G, Done), _) :- !,
	'$when'(C, G, Done).
'$redo_eq'('$done', _ ,_ , Goal, _) :-
	'$execute'(Goal).

%
% ground is similar to freeze
'$redo_ground'(Done, X, Goal) :- nonvar(Done), !.
'$redo_ground'(Done, X, Goal) :-
	'$non_ground'(X, Var), !,
	'$freeze'(Var, '$redo_ground'(Done, X, Goal)).
'$redo_ground'(Done, _, '$when'(C, G, Done)) :- !,
	'$when'(C, G, Done).
'$redo_ground'('$done', _, Goal) :-
	'$execute'(Goal).


%
% support for when/2 built-in
%
when(Conds,Goal) :-
	'$current_module'(Mod,Mod),
	'$prepare_goal_for_when'(Goal, Mod, ModG),
	'$when'(Conds, ModG, Done, [], LG), !,
%write(vsc:freezing(LG,Done)),nl,
	'$suspend_when_goals'(LG, Done).
when(_,Goal) :-
	'$execute'(Goal).

%
% support for when/2 like declaration.
%
%
% when will block on a conjunction or disjunction of nonvar, ground,
% ?=, where ?= is both terms being bound together
%
%
'$declare_when'(Cond, G) :-
	'$generate_code_for_when'(Cond, G, Code),
	'$$compile'(Code, 5), fail.
'$declare_when'(_,_).

%
% use a meta interpreter for now
%
'$generate_code_for_when'(Conds, G,
	( G :- '$when'(Conds, ModG, Done, [], LG), !,
	'$suspend_when_goals'(LG, Done)) ) :-
	'$current_module'(Mod,Mod),
	'$prepare_goal_for_when'(G, Mod, ModG).


%
% make sure we have module info for G!
%
'$prepare_goal_for_when'(G, Mod, Mod:call(G)) :- var(G), !.
'$prepare_goal_for_when'(M:G, _,  M:G) :- !.
'$prepare_goal_for_when'(G, Mod, Mod:G).
	

%
% now for the important bit
%

% Done is used to synchronise: when it is bound someone else did the
% goal and we can give up.
%
% $when/5 and $when_suspend succeds when there is need to suspend a goal
%
%
'$when'(nonvar(V), G, Done, LG0, LGF) :-
	'$when_suspend'(nonvar(V), G, Done, LG0, LGF).
'$when'(?=(X,Y), G, Done, LG0, LGF) :-
	'$when_suspend'(?=(X,Y), G, Done, LG0, LGF).
'$when'(ground(T), G, Done, LG0, LGF) :-
	'$when_suspend'(ground(T), G, Done, LG0, LGF).
'$when'((C1, C2), G, Done, LG0, LGF) :-
	% leave it open to continue with when.
	(
	    '$when'(C1, '$when'(C2, G, Done), Done, LG0, LGI)
        ->
	    LGI = LGF
        ;
	    % we solved C1, great, now we just have to solve C2!
	    '$when'(C2, G, Done, LG0, LGF)
        ).
'$when'((G1 ; G2), G, Done, LG0, LGF) :-
	'$when'(G1, G, Done, LG0, LGI),
	'$when'(G2, G, Done, LGI, LGF).

%
% Auxiliary predicate called from within a conjunction.
% Repeat basic code for when,  as inserted in first clause for predicate.
%
'$when'(_, _, Done) :-
	nonvar(Done), !.
'$when'(Cond, G, Done) :-
	'$when'(Cond, G, Done, [], LG),
	!,
	'$suspend_when_goals'(LG, Done).
'$when'(Cond, G, '$done') :-
	'$execute'(G).

%
% Do something depending on the condition!
%
% some one else did the work.
%
'$when_suspend'(_, _, Done, _, []) :- nonvar(Done), !.
%
% now for the serious stuff.
%
'$when_suspend'(nonvar(V), G, Done, LG0, LGF) :-
	'$try_freeze'(V, G, Done, LG0, LGF).
'$when_suspend'(?=(X,Y), G, Done, LG0, LGF) :-
	'$try_eq'(X, Y, G, Done, LG0, LGF).
'$when_suspend'(ground(X), G, Done, LG0, LGF) :-
	'$try_ground'(X, G, Done, LG0, LGF).


'$try_freeze'(V, G, Done, LG0, LGF) :-
	var(V),
	LGF = ['$freeze'(V, '$redo_freeze'(Done, V, G))|LG0].

'$try_eq'(X, Y, G, Done, LG0, LGF) :- 
	'$can_unify'(X, Y, LVars), LVars = [_|_],
	LGF = ['$dif_suspend_on_lvars'(LVars, '$redo_eq'(Done, X, Y, G))|LG0].

'$try_ground'(X, G, Done, LG0, LGF) :-
	'$non_ground'(X, Var),    % the C predicate that succeds if
				  % finding out the term is nonground
				  % and gives the first variable it
				  % finds. Notice that this predicate
				  % must know about svars.
	LGF = ['$freeze'(Var, '$redo_ground'(Done, X, G))| LG0].

%
% When executing a when, if nobody succeeded, we need to create suspensions.
%
'$suspend_when_goals'([], _).
'$suspend_when_goals'(['$freeze'(V, G)|Ls], Done) :-
	var(Done), !,
	'$freeze'(V, G),
	'$suspend_when_goals'(Ls, Done).
'$suspend_when_goals'(['$dif_suspend_on_lvars'(LVars, G)|LG], Done) :-
	var(Done), !,
	'$dif_suspend_on_lvars'(LVars, G),
	'$suspend_when_goals'(Ls, Done).
'$suspend_when_goals'([_|_], _).

%
% Support for wait declarations on goals.
% Or we also use the more powerful, SICStus like, "block" declarations.
%
% block or wait declarations must precede the first clause.
%

%
% I am using the simplest solution now: I'll add an extra clause at
% the beginning of the procedure to do this work. This creates a
% choicepoint and make things a bit slower, but it's probably not as
% significant as the remaining overheads.
%
block(Conds) :-
	'$generate_blocking_code'(Conds, Code),
	'$$compile'(Code, 5), fail.
block(_).

'$generate_blocking_code'(Conds, (G :- Bd)) :-
	'$extract_head_for_block'(Conds, G),
	'$generate_body_for_block'(Conds, G, Bd).

%
% find out what we are blocking on.
%
'$extract_head_for_block'((C1, C2), G) :- !,
	'$extract_head_for_block'(C1, G).
'$extract_head_for_block'(C, G) :-
	functor(C, Na, Ar),
	functor(G, Na, Ar).

%
% If we suspend on the conditions, we should continue
% execution. If we don't suspend we should fail so that we can take
% the next clause. To
% know what we have to do we just test how many variables we suspended
% on ;-).
%

'$generate_body_for_block'((C1, C2), G, (Code1 -> ! ; Code2)) :- !,
	'$generate_for_cond_in_block'(C1, G, Code1),
	'$generate_body_for_block'(C2, G, Code2).
'$generate_body_for_block'(C, G, (Code -> ! ; fail)) :-
	'$generate_for_cond_in_block'(C, G, Code).

'$generate_for_cond_in_block'(C, G, freeze_on_first(Vars, [M|G])) :-
	C =.. [_|Args],
	G =.. [_|GArgs],
	'$generate_for_each_arg_in_block'(Args, GArgs, LVars),
	Vars =.. ['$suspend'|LVars],
	'$current_module'(M, M).


'$generate_for_each_arg_in_block'([], [], []).
'$generate_for_each_arg_in_block'(['?'|Args], [_|GArgs], LV) :-
	'$generate_for_each_arg_in_block'(Args, GArgs, LV).
'$generate_for_each_arg_in_block'(['-'|Args], [GArg|GArgs], [GArg|LV]) :-
	'$generate_for_each_arg_in_block'(Args, GArgs, LV).


%
% The wait declaration is a simpler and more efficient version of block.
%
wait(Na/Ar) :-
	functor(S, Na, Ar),
	arg(1, S, A),
	'$$compile'((S :- var(A), !, freeze(A, S)), 5), fail.
wait(_).

frozen(V, G) :- nonvar(V), !, G = true.
frozen(V, G) :-
	'$frozen_goals'(V, G0),
%write(vsc:G0), nl,
	'$purge_done_goals'(G0, GI),
	'$sort'(GI, GII),
%write(vsc:GII), nl,
	'$convert_list_of_frozen_goals'(GII, G).

'$purge_done_goals'([], []).
'$purge_done_goals'([V|G0], [V|GF]) :- '$is_att_variable'(V), !,
	'$purge_done_goals'(G0, GF).
'$purge_done_goals'(['$redo_dif'(Done, _ , _)|G0], GF) :- nonvar(Done), !,
	'$purge_done_goals'(G0, GF).
'$purge_done_goals'(['$redo_freeze'(Done, _, _)|G0], GF) :- nonvar(Done), !,
	'$purge_done_goals'(G0, GF).
'$purge_done_goals'(['$redo_eq'(Done, _, _, _)|G0], GF) :- nonvar(Done), !,
	'$purge_done_goals'(G0, GF).
'$purge_done_goals'(['$redo_ground'(Done, _, _)|G0], GF) :- nonvar(Done), !,
	'$purge_done_goals'(G0, GF).
'$purge_done_goals'([G|G0], [G|GF]) :-
	'$purge_done_goals'(G0, GF).


'$convert_list_of_frozen_goals'([], true).
'$convert_list_of_frozen_goals'([A|B], G) :-
	'$convert_list_of_frozen_goals_into_conjunction'([A|B], G).

'$convert_list_of_frozen_goals_into_conjunction'([A], G) :- !,
	'$convert_frozen_goal'(A, G).
'$convert_list_of_frozen_goals_into_conjunction'([A|L], (NA,G)) :- !,
	'$convert_frozen_goal'(A, NA),
	'$convert_list_of_frozen_goals_into_conjunction'(L, G).

'$convert_frozen_goal'(V, Gs) :- '$is_att_variable'(V), !,
	attributes:convert_att_var(V, Gs).
'$convert_frozen_goal'(M:G, M:G).
'$convert_frozen_goal'('$redo_dif'(_, X, Y), dif(X,Y)).
'$convert_frozen_goal'('$redo_freeze'(_, _, G0), GF) :-
	'$convert_frozen_goal'(G0, GF).
'$convert_frozen_goal'('$redo_eq'(_, _, _, G0), GF) :-
	'$convert_frozen_goal'(G0, GF).
'$convert_frozen_goal'('$redo_ground'(_, _, G0), GF) :-
	'$convert_frozen_goal'(G0, GF).
'$convert_frozen_goal'('$when'(_,G0,_), GF) :-
	'$convert_frozen_goal'(G0, GF).

call_residue(Goal,Residue) :-
	'$read_svar_list'(OldList,OldAttsList),
	(  create_mutable([], CurrentList),
	  create_mutable([], CurrentAttsList),
          '$set_svar_list'(CurrentList, CurrentAttsList),
	  '$execute'(Goal),
	  '$call_residue_continuation'(Goal,Residue),
          ( '$set_svar_list'(OldList,OldAttsList) ; '$set_svar_list'(CurrentList,CurrentAttsList), fail )
           ;
          '$set_svar_list'(OldList,OldAttsList), fail
        ).

%
% goal needs to be in a different procedure to catch suspended goals.
%
'$call_residue_continuation'(Goal,Residue) :-
	'$project'(Goal),
	'$all_frozen_goals'(LGs0),
	sort(LGs0, LGs),
	'$purge_and_set_done_goals'(LGs, Residue, []).

'$purge_and_set_done_goals'([], L, L).
'$purge_and_set_done_goals'([AttV|G0], [LVars-GS|GF], Atts) :-
	'$is_att_variable'(AttV), !,
	attributes:convert_att_var(AttV, Gs),
	'$purge_and_set_done_goals'(G0, GF, Atts).
'$purge_and_set_done_goals'(['$redo_dif'(Done, X , Y)|G0], [LVars-dif(X,Y)|GF], Atts) :-
	var(Done), !,
	Done = '$done',
	'$can_unify'(X, Y, LVars),
	'$purge_and_set_done_goals'(G0, GF, Atts).
'$purge_and_set_done_goals'(['$redo_freeze'(Done, V, G)|G0], [[V]-freeze(G)|GF], Atts) :-
        var(Done), !,
	Done = '$done',
	'$purge_and_set_done_goals'(G0, GF, Atts).
'$purge_and_set_done_goals'(['$redo_eq'(Done, X, Y, G)|G0], [[X,Y]-G|GF], Atts) :-
	var(Done), !,
	Done = '$done',
	'$purge_and_set_done_goals'(G0, GF, Atts).
'$purge_and_set_done_goals'(['$redo_ground'(Done, X, G)|G0], [[X]-G|GF], Atts) :-
	var(Done), !,
	Done = '$done',
	'$purge_and_set_done_goals'(G0, GF, Atts).
'$purge_and_set_done_goals'([G|G0], GF, Atts) :-
	Done = '$done',
	'$purge_and_set_done_goals'(G0, GF, Atts).


'$project'(_) :-
	'$undefined'(attributes:modules_with_attributes(_)), !.

'$project'(G) :-
	attributes:modules_with_attributes(LMods),
	term_variables(G,LIV),
	'$all_attvars'(LAV),
	'$project_module'(LMods,LIV,LAV).

'$project_module'([],_,_).
'$project_module'([Mod|LMods],LIV,LAV) :-
	\+ '$undefined'(Mod:project_attributes(LIV, LAV)),
	'$execute'(Mod:project_attributes(LIV, LAV)), !,
	'$project_module'(LMods,LIV,LAV).
'$project_module'([Mod|LMods],LIV,LAV) :-
	'$project_module'(LMods,LIV,LAV).

